@@ -17,26 +17,24 @@ module Network.Wai.Handler.Warp.FdCache (
 import Control.Exception (bracket)
 import Control.Reaper
 import Data.IORef
-import Network.Wai.Handler.Warp.MultiMap
+import Network.Wai.Handler.Warp.MultiMap as MM
 import System.Posix.IO (openFd, OpenFileFlags(..), defaultFileFlags, OpenMode(ReadOnly), closeFd, FdOption(CloseOnExec), setFdOption)
 #endif
 import System.Posix.Types (Fd)
 
 ----------------------------------------------------------------
 
-type Hash = Int
-
 -- | An action to activate a Fd cache entry.
 type Refresh = IO ()
 
-getFdNothing :: Hash -> FilePath -> IO (Maybe Fd, Refresh)
-getFdNothing _ _ = return (Nothing, return ())
+getFdNothing :: FilePath -> IO (Maybe Fd, Refresh)
+getFdNothing _ = return (Nothing, return ())
 
 ----------------------------------------------------------------
 
 -- | Creating 'MutableFdCache' and executing the action in the second
 --   argument. The first argument is a cache duration in second.
-withFdCache :: Int -> ((Hash -> FilePath -> IO (Maybe Fd, Refresh)) -> IO a) -> IO a
+withFdCache :: Int -> ((FilePath -> IO (Maybe Fd, Refresh)) -> IO a) -> IO a
 #ifdef WINDOWS
 withFdCache _        action = action getFdNothing
 #else
@@ -65,7 +63,7 @@ inactive (MutableStatus ref) = writeIORef ref Inactive
 
 ----------------------------------------------------------------
 
-data FdEntry = FdEntry !FilePath !Fd !MutableStatus
+data FdEntry = FdEntry !Fd !MutableStatus
 
 openFile :: FilePath -> IO Fd
 openFile path = do
@@ -77,25 +75,23 @@ closeFile :: Fd -> IO ()
 closeFile = closeFd
 
 newFdEntry :: FilePath -> IO FdEntry
-newFdEntry path = FdEntry path <$> openFile path <*> newActiveStatus
+newFdEntry path = FdEntry <$> openFile path <*> newActiveStatus
 
 setFileCloseOnExec :: Fd -> IO ()
 setFileCloseOnExec fd = setFdOption fd CloseOnExec True
 
 ----------------------------------------------------------------
 
-type FdCache = MMap FdEntry
+type FdCache = MultiMap FdEntry
 
 -- | Mutable Fd cacher.
-newtype MutableFdCache = MutableFdCache (Reaper FdCache (Hash, FdEntry))
+newtype MutableFdCache = MutableFdCache (Reaper FdCache (FilePath,FdEntry))
 
 fdCache :: MutableFdCache -> IO FdCache
 fdCache (MutableFdCache reaper) = reaperRead reaper
 
-look :: MutableFdCache -> FilePath -> Hash -> IO (Maybe FdEntry)
-look mfc path key = searchWith key check <$> fdCache mfc
-  where
-    check (FdEntry path' _ _) = path == path'
+look :: MutableFdCache -> FilePath -> IO (Maybe FdEntry)
+look mfc path = MM.lookup path <$> fdCache mfc
 
 ----------------------------------------------------------------
 
@@ -116,7 +112,7 @@ clean old = do
     new <- pruneWith old prune
     return $ merge new
   where
-    prune (FdEntry _ fd mst) = status mst >>= act
+    prune (_,FdEntry fd mst) = status mst >>= act
       where
         act Active   = inactive mst >> return True
         act Inactive = closeFd fd   >> return False
@@ -126,21 +122,21 @@ clean old = do
 terminate :: MutableFdCache -> IO ()
 terminate (MutableFdCache reaper) = do
     !t <- reaperStop reaper
-    mapM_ closeIt $ toList t
+    mapM_ (closeIt . snd) $ toList t
   where
-    closeIt (FdEntry _ fd _) = closeFd fd
+    closeIt (FdEntry fd _) = closeFd fd
 
 ----------------------------------------------------------------
 
 -- | Getting 'Fd' and 'Refresh' from the mutable Fd cacher.
-getFd :: MutableFdCache -> Hash -> FilePath -> IO (Maybe Fd, Refresh)
-getFd mfc@(MutableFdCache reaper) h path = look mfc path h >>= get
+getFd :: MutableFdCache -> FilePath -> IO (Maybe Fd, Refresh)
+getFd mfc@(MutableFdCache reaper) path = look mfc path >>= get
   where
     get Nothing = do
-        ent@(FdEntry _ fd mst) <- newFdEntry path
-        reaperAdd reaper (h, ent)
+        ent@(FdEntry fd mst) <- newFdEntry path
+        reaperAdd reaper (path,ent)
         return (Just fd, refresh mst)
-    get (Just (FdEntry _ fd mst)) = do
+    get (Just (FdEntry fd mst)) = do
         refresh mst
         return (Just fd, refresh mst)
 #endif
