@@ -1,10 +1,9 @@
-{-# LANGUAGE OverloadedStrings #-}
-
 module Network.Wai.Handler.Warp.HTTP2.EncodeFrame where
 
-import Data.ByteString (ByteString)
 import Network.HTTP2
 
+import Network.Wai.Handler.Warp.Imports
+
 ----------------------------------------------------------------
 
 goawayFrame :: StreamId -> ErrorCodeId -> ByteString -> ByteString
