@@ -23,8 +23,9 @@ import Network.Socket (Socket, close, accept, withSocketsDo, SockAddr(SockAddrIn
 import qualified Network.Socket.ByteString as Sock
 import Network.Wai
 import Network.Wai.Internal (ResponseReceived (ResponseReceived))
-import System.Environment (getEnvironment)
+import System.Environment (lookupEnv)
 import System.Timeout (timeout)
+import qualified System.TimeManager as T
 
 import Network.Wai.Handler.Warp.Buffer
 import Network.Wai.Handler.Warp.Counter
@@ -40,7 +41,6 @@ import Network.Wai.Handler.Warp.Request
 import Network.Wai.Handler.Warp.Response
 import Network.Wai.Handler.Warp.SendFile
 import Network.Wai.Handler.Warp.Settings
-import qualified Network.Wai.Handler.Warp.Timeout as T
 import Network.Wai.Handler.Warp.Types
 
 
@@ -80,7 +80,7 @@ run p = runSettings defaultSettings { settingsPort = p }
 -- Since 3.0.9
 runEnv :: Port -> Application -> IO ()
 runEnv p app = do
-    mp <- lookup "PORT" <$> getEnvironment
+    mp <- lookupEnv "PORT"
 
     maybe (run p app) runReadPort mp
 
@@ -168,15 +168,15 @@ runSettingsConnectionMakerSecure :: Settings -> IO (IO (Connection, Transport),
 runSettingsConnectionMakerSecure set getConnMaker app = do
     settingsBeforeMainLoop set
     counter <- newCounter
-    withII0 $ acceptConnection set getConnMaker app counter
+    withII $ acceptConnection set getConnMaker app counter
   where
-    withII0 action =
+    withII action =
         withTimeoutManager $ \tm ->
         D.withDateCache $ \dc ->
         F.withFdCache fdCacheDurationInSeconds $ \fdc ->
         I.withFileInfoCache fdFileInfoDurationInSeconds $ \fic -> do
-            let ii0 = InternalInfo0 tm dc fdc fic
-            action ii0
+            let ii = InternalInfo tm dc fdc fic
+            action ii
 
     !fdCacheDurationInSeconds = settingsFdCacheDuration set * 1000000
     !fdFileInfoDurationInSeconds = settingsFileInfoCacheDuration set * 1000000
@@ -205,9 +205,9 @@ acceptConnection :: Settings
                  -> IO (IO (Connection, Transport), SockAddr)
                  -> Application
                  -> Counter
-                 -> InternalInfo0
+                 -> InternalInfo
                  -> IO ()
-acceptConnection set getConnMaker app counter ii0 = do
+acceptConnection set getConnMaker app counter ii = do
     -- First mask all exceptions in acceptLoop. This is necessary to
     -- ensure that no async exception is throw between the call to
     -- acceptNewConnection and the registering of connClose.
@@ -234,7 +234,7 @@ acceptConnection set getConnMaker app counter ii0 = do
         case mx of
             Nothing             -> return ()
             Just (mkConn, addr) -> do
-                fork set mkConn addr app counter ii0
+                fork set mkConn addr app counter ii
                 acceptLoop
 
     acceptNewConnection = do
@@ -257,9 +257,9 @@ fork :: Settings
      -> SockAddr
      -> Application
      -> Counter
-     -> InternalInfo0
+     -> InternalInfo
      -> IO ()
-fork set mkConn addr app counter ii0 = settingsFork set $ \unmask ->
+fork set mkConn addr app counter ii = settingsFork set $ \unmask ->
     -- Call the user-supplied on exception code if any
     -- exceptions are thrown.
     handle (settingsOnException set Nothing) .
@@ -291,9 +291,8 @@ fork set mkConn addr app counter ii0 = settingsFork set $ \unmask ->
     -- the connection immediately in case the child thread catches the
     -- async exception or performs some long-running cleanup action.
     serve unmask ref (conn, transport) = bracket register cancel $ \th -> do
-        let ii1 = toInternalInfo1 ii0 th
         -- We now have fully registered a connection close handler in
-        -- the case of all exceptions, so it is safe to one again
+        -- the case of all exceptions, so it is safe to once again
         -- allow async exceptions.
         unmask .
             -- Call the user-supplied code for connection open and
@@ -301,9 +300,9 @@ fork set mkConn addr app counter ii0 = settingsFork set $ \unmask ->
            bracket (onOpen addr) (onClose addr) $ \goingon ->
            -- Actually serve this connection.  bracket with closeConn
            -- above ensures the connection is closed.
-           when goingon $ serveConnection conn ii1 addr transport set app
+           when goingon $ serveConnection conn ii th addr transport set app
       where
-        register = T.registerKillThread (timeoutManager0 ii0)
+        register = T.registerKillThread (timeoutManager ii)
                                         (closeConn ref conn)
         cancel   = T.cancel
 
@@ -311,13 +310,14 @@ fork set mkConn addr app counter ii0 = settingsFork set $ \unmask ->
     onClose adr _ = decrease counter >> settingsOnClose set adr
 
 serveConnection :: Connection
-                -> InternalInfo1
+                -> InternalInfo
+                -> T.Handle
                 -> SockAddr
                 -> Transport
                 -> Settings
                 -> Application
                 -> IO ()
-serveConnection conn ii1 origAddr transport settings app = do
+serveConnection conn ii th origAddr transport settings app = do
     -- fixme: Upgrading to HTTP/2 should be supported.
     (h2,bs) <- if isHTTP2 transport then
                    return (True, "")
@@ -332,7 +332,7 @@ serveConnection conn ii1 origAddr transport settings app = do
         rawRecvN <- makeReceiveN bs (connRecv conn) (connRecvBuf conn)
         let recvN = wrappedRecvN th istatus (settingsSlowlorisSize settings) rawRecvN
         -- fixme: origAddr
-        http2 conn ii1 origAddr transport settings recvN app
+        http2 conn ii origAddr transport settings recvN app
       else do
         src <- mkSource (wrappedRecv conn th istatus (settingsSlowlorisSize settings))
         writeIORef istatus True
@@ -344,7 +344,7 @@ serveConnection conn ii1 origAddr transport settings app = do
             -- https://github.com/yesodweb/wai/issues/618
             Just NoKeepAliveRequest -> return ()
             Nothing -> do
-              sendErrorResponse (dummyreq addr) istatus e
+              _ <- sendErrorResponse (dummyreq addr) istatus e
               throwIO e
 
   where
@@ -388,29 +388,26 @@ serveConnection conn ii1 origAddr transport settings app = do
 
     decodeAscii = map (chr . fromEnum) . S.unpack
 
-    th = threadHandle1 ii1
-
     shouldSendErrorResponse se
         | Just ConnectionClosedByPeer <- fromException se = False
         | otherwise                                       = True
 
     sendErrorResponse req istatus e = do
         status <- readIORef istatus
-        when (shouldSendErrorResponse e && status) $ do
-           let ii = toInternalInfo ii1 0 -- dummy
-           void $ sendResponse settings conn ii req defaultIndexRequestHeader (return S.empty) (errorResponse e)
+        if shouldSendErrorResponse e && status
+            then do
+                sendResponse settings conn ii th req defaultIndexRequestHeader (return S.empty) (errorResponse e)
+            else return False
 
     dummyreq addr = defaultRequest { remoteHost = addr }
 
     errorResponse e = settingsOnExceptionResponse settings e
 
     http1 firstRequest addr istatus src = do
-        (req', mremainingRef, idxhdr, nextBodyFlush, ii) <- recvRequest firstRequest settings conn ii1 addr src
+        (req', mremainingRef, idxhdr, nextBodyFlush) <- recvRequest firstRequest settings conn ii th addr src
         let req = req' { isSecure = isTransportSecure transport }
-        keepAlive <- processRequest istatus src req mremainingRef idxhdr nextBodyFlush ii
+        keepAlive <- processRequest istatus src req mremainingRef idxhdr nextBodyFlush
             `E.catch` \e -> do
-                -- Call the user-supplied exception handlers, passing the request.
-                sendErrorResponse req istatus e
                 settingsOnException settings (Just req) e
                 -- Don't throw the error again to prevent calling settingsOnException twice.
                 return False
@@ -425,7 +422,7 @@ serveConnection conn ii1 origAddr transport settings app = do
         -- and ignore. See: https://github.com/yesodweb/wai/issues/618
         when keepAlive $ http1 False addr istatus src
 
-    processRequest istatus src req mremainingRef idxhdr nextBodyFlush ii = do
+    processRequest istatus src req mremainingRef idxhdr nextBodyFlush = do
         -- Let the application run for as long as it wants
         T.pause th
 
@@ -433,30 +430,38 @@ serveConnection conn ii1 origAddr transport settings app = do
         -- creating the request, we need to make sure that we don't get
         -- an async exception before calling the ResponseSource.
         keepAliveRef <- newIORef $ error "keepAliveRef not filled"
-        _ <- app req $ \res -> do
+        r <- E.try $ app req $ \res -> do
             T.resume th
             -- FIXME consider forcing evaluation of the res here to
             -- send more meaningful error messages to the user.
             -- However, it may affect performance.
             writeIORef istatus False
-            keepAlive <- sendResponse settings conn ii req idxhdr (readSource src) res
+            keepAlive <- sendResponse settings conn ii th req idxhdr (readSource src) res
             writeIORef keepAliveRef keepAlive
             return ResponseReceived
+        case r of
+            Right ResponseReceived -> return ()
+            Left e@(SomeException _)
+              | Just (ExceptionInsideResponseBody e') <- fromException e -> throwIO e'
+              | otherwise -> do
+                    keepAlive <- sendErrorResponse req istatus e
+                    settingsOnException settings (Just req) e
+                    writeIORef keepAliveRef keepAlive
+
         keepAlive <- readIORef keepAliveRef
 
         -- We just send a Response and it takes a time to
         -- receive a Request again. If we immediately call recv,
-        -- it is likely to fail and the IO manager works.
-        -- It is very costly. So, we yield to another Haskell
+        -- it is likely to fail and cause the IO manager to do some work.
+        -- It is very costly, so we yield to another Haskell
         -- thread hoping that the next Request will arrive
         -- when this Haskell thread will be re-scheduled.
         -- This improves performance at least when
         -- the number of cores is small.
         Conc.yield
 
-        if not keepAlive then
-            return False
-          else
+        if keepAlive
+          then
             -- If there is an unknown or large amount of data to still be read
             -- from the request body, simple drop this connection instead of
             -- reading it all in to satisfy a keep-alive request.
@@ -482,6 +487,8 @@ serveConnection conn ii1 origAddr transport settings app = do
                               else
                                 return False
                         Nothing -> tryKeepAlive
+          else
+            return False
 
 flushEntireBody :: IO ByteString -> IO ()
 flushEntireBody src =
@@ -536,7 +543,13 @@ setSocketCloseOnExec :: Socket -> IO ()
 #if WINDOWS
 setSocketCloseOnExec _ = return ()
 #else
-setSocketCloseOnExec socket = F.setFileCloseOnExec $ fromIntegral $ fdSocket socket
+setSocketCloseOnExec socket = do
+#if MIN_VERSION_network(3,0,0)
+    fd <- fdSocket socket
+#else
+    let fd = fdSocket socket
+#endif
+    F.setFileCloseOnExec $ fromIntegral fd
 #endif
 
 gracefulShutdown :: Settings -> Counter -> IO ()
