@@ -10,12 +10,12 @@ import Data.IORef (IORef, readIORef, writeIORef, newIORef)
 import Data.Typeable (Typeable)
 import Foreign.Ptr (Ptr)
 import System.Posix.Types (Fd)
+import qualified System.TimeManager as T
 
 import qualified Network.Wai.Handler.Warp.Date as D
 import qualified Network.Wai.Handler.Warp.FdCache as F
 import qualified Network.Wai.Handler.Warp.FileInfoCache as I
 import Network.Wai.Handler.Warp.Imports
-import qualified Network.Wai.Handler.Warp.Timeout as T
 
 ----------------------------------------------------------------
 
@@ -52,6 +52,20 @@ instance Exception InvalidRequest
 
 ----------------------------------------------------------------
 
+-- | Exception thrown if something goes wrong while in the midst of
+-- sending a response, since the status code can't be altered at that
+-- point.
+--
+-- Used to determine whether keeping the HTTP1.1 connection / HTTP2 stream alive is safe
+-- or irrecoverable.
+
+newtype ExceptionInsideResponseBody = ExceptionInsideResponseBody SomeException
+    deriving (Show, Typeable)
+
+instance Exception ExceptionInsideResponseBody
+
+----------------------------------------------------------------
+
 -- | Data type to abstract file identifiers.
 --   On Unix, a file descriptor would be specified to make use of
 --   the file descriptor cache.
@@ -112,42 +126,13 @@ data Connection = Connection {
 
 ----------------------------------------------------------------
 
-type Hash = Int
-
-
-data InternalInfo0 =
-    InternalInfo0 T.Manager
-                  (IO D.GMTDate)
-                  (Hash -> FilePath -> IO (Maybe F.Fd, F.Refresh))
-                  (Hash -> FilePath -> IO I.FileInfo)
-
-timeoutManager0 :: InternalInfo0 -> T.Manager
-timeoutManager0 (InternalInfo0 tm _ _ _) = tm
-
-data InternalInfo1 =
-    InternalInfo1 T.Handle
-                  T.Manager
-                  (IO D.GMTDate)
-                  (Hash -> FilePath -> IO (Maybe F.Fd, F.Refresh))
-                  (Hash -> FilePath -> IO I.FileInfo)
-
-toInternalInfo1 :: InternalInfo0 -> T.Handle -> InternalInfo1
-toInternalInfo1 (InternalInfo0 b c d e) a = InternalInfo1 a b c d e
-
-threadHandle1 :: InternalInfo1 -> T.Handle
-threadHandle1 (InternalInfo1 th _ _ _ _) = th
-
 data InternalInfo = InternalInfo {
-    threadHandle   :: T.Handle
-  , timeoutManager :: T.Manager
+    timeoutManager :: T.Manager
   , getDate        :: IO D.GMTDate
   , getFd          :: FilePath -> IO (Maybe F.Fd, F.Refresh)
   , getFileInfo    :: FilePath -> IO I.FileInfo
   }
 
-toInternalInfo :: InternalInfo1 -> Hash -> InternalInfo
-toInternalInfo (InternalInfo1 a b c d e) h = InternalInfo a b c (d h) (e h)
-
 ----------------------------------------------------------------
 
 -- | Type for input streaming.
