@@ -71,11 +71,11 @@ spec = do
 headers :: H.ResponseHeaders
 headers = [
     ("Date", "Mon, 13 Aug 2012 04:22:55 GMT")
-  , ("Content-Lenght", "151")
+  , ("Content-Length", "151")
   , ("Server", "Mighttpd/2.5.8")
   , ("Last-Modified", "Fri, 22 Jun 2012 01:18:08 GMT")
   , ("Content-Type", "text/html")
   ]
 
 composedHeader :: ByteString
-composedHeader = "HTTP/1.1 200 OK\r\nDate: Mon, 13 Aug 2012 04:22:55 GMT\r\nContent-Lenght: 151\r\nServer: Mighttpd/2.5.8\r\nLast-Modified: Fri, 22 Jun 2012 01:18:08 GMT\r\nContent-Type: text/html\r\n\r\n"
+composedHeader = "HTTP/1.1 200 OK\r\nDate: Mon, 13 Aug 2012 04:22:55 GMT\r\nContent-Length: 151\r\nServer: Mighttpd/2.5.8\r\nLast-Modified: Fri, 22 Jun 2012 01:18:08 GMT\r\nContent-Type: text/html\r\n\r\n"
