@@ -1,16 +1,16 @@
-{-# LANGUAGE CPP #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 
-module RunSpec (main, spec, withApp, connectTo) where
+module RunSpec (main, spec, withApp, connectTo, MySocket, msWrite, msRead, msClose) where
 
 import Control.Concurrent (forkIO, killThread, threadDelay)
 import Control.Concurrent.MVar (newEmptyMVar, takeMVar, putMVar)
+import Control.Concurrent.STM
 import qualified Control.Exception as E
 import Control.Exception.Lifted (bracket, try, IOException, onException)
 import Control.Monad (forM_, replicateM_, unless)
 import Control.Monad.IO.Class (MonadIO, liftIO)
-import Data.ByteString (ByteString, hPutStr, hGetSome)
+import Data.ByteString (ByteString)
 import qualified Data.ByteString as S
 import Data.ByteString.Builder (byteString)
 import qualified Data.ByteString.Char8 as S8
@@ -21,8 +21,8 @@ import Network.HTTP.Types
 import Network.Socket
 import Network.Socket.ByteString (sendAll)
 import Network.Wai hiding (responseHeaders)
+import Network.Wai.Internal (getRequestBodyChunk)
 import Network.Wai.Handler.Warp
-import System.IO (hFlush, hClose, Handle, IOMode(..))
 import System.IO.Unsafe (unsafePerformIO)
 import System.Timeout (timeout)
 import Test.Hspec
@@ -35,13 +35,44 @@ main = hspec spec
 type Counter = I.IORef (Either String Int)
 type CounterApplication = Counter -> Application
 
-connectTo :: HostName -> Int -> IO Handle
-connectTo host port = do
-    let hints = defaultHints { addrSocketType = Stream }
-    addr:_ <- getAddrInfo (Just hints) (Just host) (Just $ show port)
-    sock <- socket (addrFamily addr) (addrSocketType addr) (addrProtocol addr)
-    connect sock $ addrAddress addr
-    socketToHandle sock ReadWriteMode
+data MySocket = MySocket
+  { msSocket :: !Socket
+  , _msBuffer :: !(I.IORef ByteString)
+  }
+
+msWrite :: MySocket -> ByteString -> IO ()
+msWrite ms bs = sendAll (msSocket ms) bs
+
+msRead :: MySocket -> Int -> IO ByteString
+msRead (MySocket s ref) expected = do
+  bs <- I.readIORef ref
+  inner (bs:) (S.length bs)
+  where
+    inner front total =
+      case compare total expected of
+        EQ -> do
+          I.writeIORef ref mempty
+          pure $ S.concat $ front []
+        GT -> do
+          let bs = S.concat $ front []
+              (x, y) = S.splitAt expected bs
+          I.writeIORef ref y
+          pure x
+        LT -> do
+          bs <- safeRecv s 4096
+          if S.null bs
+            then do
+              I.writeIORef ref mempty
+              pure $ S.concat $ front []
+            else inner (front . (bs:)) (total + S.length bs)
+
+msClose :: MySocket -> IO ()
+msClose = Network.Socket.close . msSocket
+
+connectTo :: Int -> IO MySocket
+connectTo port = MySocket
+  <$> (fst <$> getSocketTCP "127.0.0.1" port)
+  <*> I.newIORef mempty
 
 incr :: MonadIO m => Counter -> m ()
 incr icount = liftIO $ I.atomicModifyIORef icount $ \ecount ->
@@ -54,7 +85,7 @@ err icount msg = liftIO $ I.writeIORef icount $ Left $ show msg
 
 readBody :: CounterApplication
 readBody icount req f = do
-    body <- consumeBody $ requestBody req
+    body <- consumeBody $ getRequestBodyChunk req
     case () of
         ()
             | pathInfo req == ["hello"] && L.fromChunks body /= "Hello"
@@ -75,8 +106,8 @@ ignoreBody icount req f = do
 
 doubleConnect :: CounterApplication
 doubleConnect icount req f = do
-    _ <- consumeBody $ requestBody req
-    _ <- consumeBody $ requestBody req
+    _ <- consumeBody $ getRequestBodyChunk req
+    _ <- consumeBody $ getRequestBodyChunk req
     incr icount
     f $ responseLBS status200 [] "double connect"
 
@@ -105,7 +136,13 @@ withApp settings app f = do
     bracket
         (forkIO $ runSettings settings' app `onException` putMVar baton ())
         killThread
-        (const $ takeMVar baton >> f port)
+        (const $ do
+            takeMVar baton
+            -- use timeout to make sure we don't take too long
+            mres <- timeout (60 * 1000 * 1000) (f port)
+            case mres of
+              Nothing -> error "Timeout triggered, too slow!"
+              Just a -> pure a)
 
 runTest :: Int -- ^ expected number of requests
         -> CounterApplication
@@ -114,9 +151,9 @@ runTest :: Int -- ^ expected number of requests
 runTest expected app chunks = do
     ref <- I.newIORef (Right 0)
     withApp defaultSettings (app ref) $ \port -> do
-        handle <- connectTo "127.0.0.1" port
-        forM_ chunks $ \chunk -> hPutStr handle chunk >> hFlush handle
-        _ <- timeout 100000 $ replicateM_ expected $ hGetSome handle 4096
+        ms <- connectTo port
+        forM_ chunks $ \chunk -> msWrite ms chunk
+        _ <- timeout 100000 $ replicateM_ expected $ msRead ms 4096
         res <- I.readIORef ref
         case res of
             Left s -> error s
@@ -132,10 +169,9 @@ runTerminateTest expected input = do
     ref <- I.newIORef Nothing
     let onExc _ = I.writeIORef ref . Just
     withApp (setOnException onExc defaultSettings) dummyApp $ \port -> do
-        handle <- connectTo "127.0.0.1" port
-        hPutStr handle input
-        hFlush handle
-        hClose handle
+        handle <- connectTo port
+        msWrite handle input
+        msClose handle
         threadDelay 1000
         res <- I.readIORef ref
         show res `shouldBe` show (Just expected)
@@ -207,13 +243,12 @@ spec = do
                     liftIO $ I.writeIORef iheaders $ requestHeaders req
                     f $ responseLBS status200 [] ""
             withApp defaultSettings app $ \port -> do
-                handle <- connectTo "127.0.0.1" port
+                handle <- connectTo port
                 let input = S.concat
                         [ "GET / HTTP/1.1\r\nfoo:    bar\r\n baz\r\n\tbin\r\n\r\n"
                         ]
-                hPutStr handle input
-                hFlush handle
-                hClose handle
+                msWrite handle input
+                msClose handle
                 threadDelay 1000
                 headers <- I.readIORef iheaders
                 headers `shouldBe`
@@ -225,13 +260,12 @@ spec = do
                     liftIO $ I.writeIORef iheaders $ requestHeaders req
                     f $ responseLBS status200 [] ""
             withApp defaultSettings app $ \port -> do
-                handle <- connectTo "127.0.0.1" port
+                handle <- connectTo port
                 let input = S.concat
                         [ "GET / HTTP/1.1\r\nfoo:bar\r\n\r\n"
                         ]
-                hPutStr handle input
-                hFlush handle
-                hClose handle
+                msWrite handle input
+                msClose handle
                 threadDelay 1000
                 headers <- I.readIORef iheaders
                 headers `shouldBe`
@@ -240,23 +274,26 @@ spec = do
 
     describe "chunked bodies" $ do
         it "works" $ do
+            countVar <- newTVarIO (0 :: Int)
             ifront <- I.newIORef id
             let app req f = do
-                    bss <- consumeBody $ requestBody req
+                    bss <- consumeBody $ getRequestBodyChunk req
                     liftIO $ I.atomicModifyIORef ifront $ \front -> (front . (S.concat bss:), ())
+                    atomically $ modifyTVar countVar (+ 1)
                     f $ responseLBS status200 [] ""
             withApp defaultSettings app $ \port -> do
-                handle <- connectTo "127.0.0.1" port
+                handle <- connectTo port
                 let input = S.concat
                         [ "POST / HTTP/1.1\r\nTransfer-Encoding: Chunked\r\n\r\n"
                         , "c\r\nHello World\n\r\n3\r\nBye\r\n0\r\n\r\n"
                         , "POST / HTTP/1.1\r\nTransfer-Encoding: Chunked\r\n\r\n"
                         , "b\r\nHello World\r\n0\r\n\r\n"
                         ]
-                hPutStr handle input
-                hFlush handle
-                hClose handle
-                threadDelay 1000
+                msWrite handle input
+                msClose handle
+                atomically $ do
+                  count <- readTVar countVar
+                  check $ count == 2
                 front <- I.readIORef ifront
                 front [] `shouldBe`
                     [ "Hello World\nBye"
@@ -264,38 +301,48 @@ spec = do
                     ]
         it "lots of chunks" $ do
             ifront <- I.newIORef id
+            countVar <- newTVarIO (0 :: Int)
             let app req f = do
-                    bss <- consumeBody $ requestBody req
+                    bss <- consumeBody $ getRequestBodyChunk req
                     I.atomicModifyIORef ifront $ \front -> (front . (S.concat bss:), ())
+                    atomically $ modifyTVar countVar (+ 1)
                     f $ responseLBS status200 [] ""
             withApp defaultSettings app $ \port -> do
-                handle <- connectTo "127.0.0.1" port
+                handle <- connectTo port
                 let input = concat $ replicate 2 $
                         ["POST / HTTP/1.1\r\nTransfer-Encoding: Chunked\r\n\r\n"] ++
                         (replicate 50 "5\r\n12345\r\n") ++
                         ["0\r\n\r\n"]
-                mapM_ (\bs -> hPutStr handle bs >> hFlush handle) input
-                hClose handle
-                threadDelay 100000 -- FIXME why does this delay need to be so high?
+                mapM_ (msWrite handle) input
+                msClose handle
+                atomically $ do
+                  count <- readTVar countVar
+                  check $ count == 2
                 front <- I.readIORef ifront
                 front [] `shouldBe` replicate 2 (S.concat $ replicate 50 "12345")
+-- For some reason, the following test on Windows causes the socket
+-- to be killed prematurely. Worth investigating in the future if possible.
         it "in chunks" $ do
             ifront <- I.newIORef id
+            countVar <- newTVarIO (0 :: Int)
             let app req f = do
-                    bss <- consumeBody $ requestBody req
+                    bss <- consumeBody $ getRequestBodyChunk req
                     liftIO $ I.atomicModifyIORef ifront $ \front -> (front . (S.concat bss:), ())
+                    atomically $ modifyTVar countVar (+ 1)
                     f $ responseLBS status200 [] ""
             withApp defaultSettings app $ \port -> do
-                handle <- connectTo "127.0.0.1" port
+                handle <- connectTo port
                 let input = S.concat
                         [ "POST / HTTP/1.1\r\nTransfer-Encoding: Chunked\r\n\r\n"
                         , "c\r\nHello World\n\r\n3\r\nBye\r\n0\r\n"
                         , "POST / HTTP/1.1\r\nTransfer-Encoding: Chunked\r\n\r\n"
                         , "b\r\nHello World\r\n0\r\n\r\n"
                         ]
-                mapM_ (\bs -> hPutStr handle bs >> hFlush handle) $ map S.singleton $ S.unpack input
-                hClose handle
-                threadDelay 1000
+                mapM_ (msWrite handle) $ map S.singleton $ S.unpack input
+                msClose handle
+                atomically $ do
+                  count <- readTVar countVar
+                  check $ count == 2
                 front <- I.readIORef ifront
                 front [] `shouldBe`
                     [ "Hello World\nBye"
@@ -304,7 +351,7 @@ spec = do
         it "timeout in request body" $ do
             ifront <- I.newIORef id
             let app req f = do
-                    bss <- (consumeBody $ requestBody req) `onException`
+                    bss <- (consumeBody $ getRequestBodyChunk req) `onException`
                         liftIO (I.atomicModifyIORef ifront (\front -> (front . ("consume interrupted":), ())))
                     liftIO $ threadDelay 4000000 `E.catch` \e -> do
                         I.atomicModifyIORef ifront (\front ->
@@ -317,16 +364,16 @@ spec = do
                 let bs1 = S.replicate 2048 88
                     bs2 = "This is short"
                     bs = S.append bs1 bs2
-                handle <- connectTo "127.0.0.1" port
-                hPutStr handle "POST / HTTP/1.1\r\n"
-                hPutStr handle "content-length: "
-                hPutStr handle $ S8.pack $ show $ S.length bs
-                hPutStr handle "\r\n\r\n"
+                handle <- connectTo port
+                msWrite handle "POST / HTTP/1.1\r\n"
+                msWrite handle "content-length: "
+                msWrite handle $ S8.pack $ show $ S.length bs
+                msWrite handle "\r\n\r\n"
                 threadDelay 100000
-                hPutStr handle bs1
+                msWrite handle bs1
                 threadDelay 100000
-                hPutStr handle bs2
-                hClose handle
+                msWrite handle bs2
+                msClose handle
                 threadDelay 5000000
                 front <- I.readIORef ifront
                 S.concat (front []) `shouldBe` bs
@@ -343,37 +390,42 @@ spec = do
                         loop
                 doubleBS = S.concatMap $ \w -> S.pack [w, w]
             withApp defaultSettings app $ \port -> do
-                handle <- connectTo "127.0.0.1" port
-                hPutStr handle "POST / HTTP/1.1\r\n\r\n12345"
-                hFlush handle
-                timeout 100000 (S.hGet handle 10) >>= (`shouldBe` Just "1122334455")
-                hPutStr handle "67890"
-                hFlush handle
-                timeout 100000 (S.hGet handle 10) >>= (`shouldBe` Just "6677889900")
-
+                ms <- connectTo port
+                msWrite ms "POST / HTTP/1.1\r\n\r\n12345"
+                timeout 100000 (msRead ms 10) >>= (`shouldBe` Just "1122334455")
+                msWrite ms "67890"
+                timeout 100000 (msRead ms 10) >>= (`shouldBe` Just "6677889900")
     it "only one date and server header" $ do
         let app _ f = f $ responseLBS status200
                 [ ("server", "server")
                 , ("date", "date")
                 ] ""
+            getValues key = map snd
+                          . filter (\(key', _) -> key == key')
+                          . responseHeaders
         withApp defaultSettings app $ \port -> do
             res <- sendGET $ "http://127.0.0.1:" ++ show port
-            getHeaderValue hServer (responseHeaders res) `shouldBe` Just "server"
-            getHeaderValue hDate (responseHeaders res) `shouldBe` Just "date"
+            getValues hServer res `shouldBe` ["server"]
+            getValues hDate res `shouldBe` ["date"]
 
     it "streaming echo #249" $ do
+        countVar <- newTVarIO (0 :: Int)
         let app req f = f $ responseStream status200 [] $ \write _ -> do
-            let loop = do
-                    bs <- requestBody req
+             let loop = do
+                    bs <- getRequestBodyChunk req
                     unless (S.null bs) $ do
                         write $ byteString bs
+                        atomically $ modifyTVar countVar (+ 1)
                         loop
-            loop
+             loop
         withApp defaultSettings app $ \port -> do
             (sock, _addr) <- getSocketTCP "127.0.0.1" port
             sendAll sock "POST / HTTP/1.1\r\ntransfer-encoding: chunked\r\n\r\n"
             threadDelay 10000
             sendAll sock "5\r\nhello\r\n0\r\n\r\n"
+            atomically $ do
+              count <- readTVar countVar
+              check $ count >= 1
             bs <- safeRecv sock 4096
             S.takeWhile (/= 13) bs `shouldBe` "HTTP/1.1 200 OK"
 
